/* 
 * LAMP - Linux and Managed Power
 * simple CPU power-management using Linux cpuFreq
 *
 * Vivek Haldar
 * Aug 2003
 *
 * License: GPL
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>
#include "lamp.h"

#define PROC_FILE_SIZE 1024


static int current_cpu_speed;

void init() 
{
	enable_userspace_governor();
	current_cpu_speed = getspeed();
}

void enable_userspace_governor()
{
	FILE *fd;
	char filename[512];
	char gov[128];
	int cpu_speed;
	int len;
	strcpy(filename, CPU_FREQ_PATH);
	strcat(filename, "scaling_governor");
	//printf("SET: Opening %s\n", filename);
	fd = fopen(filename, "w");
	sprintf(gov, "%s", "userspace");
	len = strlen(gov);
	//printf("SET: writing %s, of len %d\n", speed, len);
	fwrite((char *)gov, 1, len, fd);
	fclose(fd);

}

void setspeed(int newspeed)
{
	FILE *fd;
	char filename[512];
	char speed[128];
	int cpu_speed;
	int len;
	strcpy(filename, CPU_FREQ_PATH);
	strcat(filename, "scaling_setspeed");
	//printf("SET: Opening %s\n", filename);
	fd = fopen(filename, "w");
	sprintf(speed, "%d", newspeed);
	len = strlen(speed);
	//printf("SET: writing %s, of len %d\n", speed, len);
	fwrite((char *)speed, 1, len, fd);
	printf("***** Changed speed to %d\n", newspeed);
	fclose(fd);
}

int getspeed()
{
	FILE *fd;
	char filename[512];
	char speed[128];
	int cpu_speed;
	strcpy(filename, CPU_FREQ_PATH);
	strcat(filename, "scaling_setspeed");
	printf("Opening %s\n", filename);
	fd = fopen(filename, "r");
	fscanf(fd, "%s", (char *)speed);
	printf("The speed was: %s\n", speed);
	cpu_speed = atoi(speed);
	fclose(fd);
	return cpu_speed;
}

void handler(int sig)
{
	int i;
	printf("signal handler for %d\n",sig);
	//showload();
	decide_switch();
}

void settimer() {
	struct itimerval rttimer;
	struct itimerval old_rttimer;

	signal(SIGALRM,handler);
	rttimer.it_value.tv_sec     = 1; /* A signal will be sent one second*/
	rttimer.it_value.tv_usec    = 0; /*  from when this is called */ 
	rttimer.it_interval.tv_sec  = 1; /* If the timer is not reset, the */
	rttimer.it_interval.tv_usec = 0; /*  signal will be sent every second */

	setitimer (ITIMER_REAL, &rttimer, &old_rttimer);
}



int
main (void)
{
	init();
	settimer();

	while (1) {
		pause();
	}

	exit (0);
}

void showload()
{
	printf ("(/proc/stat file) CPU load = %5.1lf %\n", calc_cpu_loading ());
}

/* 
 * this decides whether we switch up or down
 */
void decide_switch()
{
	double load;
	int speed_now;

	speed_now = getspeed();
	load = calc_cpu_loading();
	printf("Current load is %5.1lf %\n", load);
	printf("Current speed is %d\n", speed_now);
	if ((load > CPU_LOAD_THRESHOLD) && (speed_now != CPU_HIGH_SPEED)) {
	       	setspeed(CPU_HIGH_SPEED); 
		current_cpu_speed = CPU_HIGH_SPEED;
	}
	if ((load < CPU_LOAD_THRESHOLD) && (speed_now != CPU_LOW_SPEED)) {
	       	setspeed(CPU_LOW_SPEED); 
		current_cpu_speed = CPU_LOW_SPEED;
	}
	printf("Current speed is %d\n", getspeed());
}

/*
 * This code is due to  Tony D. (tonyd@sensis.com)
 * from comp.os.linux.development.apps
 */
static char buf[PROC_FILE_SIZE];	/*minimize stack size */
static unsigned long prev_idle = 0;
static double prev_total = 0.0;
/*uses the /proc/stat file method*/
	double
calc_cpu_loading (void)
{

	FILE *fPtr = NULL;
	char *item = NULL, *arg = NULL;
	unsigned long delta_idle = 0;
	unsigned long user = 0, nice = 0, idle = 0, sys = 0;
	double total = 0.0, time_passed = 0.0;

	/*default return value means error occurred */
	double percentCpuLoad = -1.0;

	if ((fPtr = fopen (PROC_STAT_FILE, "r")) == NULL)
		return (percentCpuLoad);

	/* read in the CPU stats that the Linux kernel has provided for us */
	/*read and analyze one line at a time from the /proc/stats file
	  and quit when we've gotten the items we need */
	while ((fgets (buf, sizeof (buf), fPtr)) != NULL)
	{
		/*get ptrs to item name and first arg in the line */
		item = strtok (buf, " \t\n");
		arg = strtok (NULL, "\n");
		if (item == NULL || arg == NULL)
			continue;
		if (item[0] == 'c' && strlen (item) == 3)
		{
			/*extract & convert to ulongs, the times we need
			  starting at the byte pointed to by arg */
			sscanf (arg, "%lu %lu %lu %lu", &user, &nice, &sys, &idle);
			break;		/*no need to read into the file further */
		}
	}				/*end of fgets while() */
	fclose (fPtr);

	/*compute time spent idling during this period */
	delta_idle = idle - prev_idle;

	/*compute time period */
	total = (double) (user + nice + sys + idle);
	time_passed = total - prev_total;

	/*Compute the % load from the time spent idling and the total time
	  in the smapling period */
	percentCpuLoad = 100.0 - (delta_idle / time_passed * 100.0);

	/*retain the current counts for next time period */
	prev_idle = idle;
	prev_total = total;

	return (percentCpuLoad);
}
